#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# 
#
# ipmicommon -- Common functions for IPMI features
#
# Doug White

import exceptions
import md5
import socket
import struct

# Some constants.
IPMI_AUTH_CALLBACK = 1
IPMI_AUTH_USER = 2
IPMI_AUTH_OPERATOR = 3
IPMI_AUTH_ADMIN = 4
IPMI_AUTH_OEM = 5

IPMI_AUTHT_NONE = 0
IPMI_AUTHT_MD2 = 1
IPMI_AUTHT_MD5 = 2
IPMI_AUTHT_PASSWD = 4
IPMI_AUTHT_OEM = 5

# Build a generic RMCP header for encapsulating IPMI messages in.
# The data is largely static.  
# The 0xff byte is the RMCP seq#, unused in IPMI.
# The last byte is the packet type:
# 0x6 for RMCP Ping
# 0x07 for IPMI Encap.
def build_rmcp_packet_header():
    return "%c%c%c%c" % (0x06, 0x00, 0xff, 0x07)
    
# IPMI Session: 
# Abstraction for handling IPMI Sessions, which maintain
# connection-oriented data.
class IPMISession:
    def __init__(self, peer, username='', password='', authlevel=IPMI_AUTH_ADMIN):
        # Create object with target pointed at peer
        self.peer = peer
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(self.sock.getsockname())
        self.sessid = 0
        self.seq = 0
        self.username = username
        self.password = password
        self.authtypes = []
        self.authlevel = authlevel
        self.authtype = None
        self.authcode = None
        self.connected = 0

    # Calculate IPMI checksum.
    def calc_ipmi_cksum(self, body):
        # This is calculated by:
        # 2's complement cksum of preceeding bytes
        # 8 bit cksum
        # for byte in bytes:
        #  cksum = (cksum+byte) % 256
        # cksum = -cksum
        # then cksum + newcksum == 0
        cksum = 0
        for x in body:
            cksum = (cksum+ord(x)) % 256
            #print "Adding %d to get %d" % (ord(x), cksum)
        # XXX how can it be modulo 256 (unsigned), yet be signed?
        #print "cksum: %d vs %d?" % (cksum, ~cksum+1)
        return -cksum
    
    def build_ipmi_channel_header(self, authtype, seq, sid, mac, leng):
        # All multibyte IPMI data is sent in Intel byte order, e.g. little endian.
        # No need to use hton* here. (bad Intel, no cookie)
        # byte 1: auth type
        # byte 2: sequence # (4 bytes)
        # byte 6: session ID (4 bytes)
        # byte 10: MAC (16 bytes)
        # there's a 1 byte Length after all this but we don't know that
        # MAC codes defined as:
        # 0 = none (when None is set AUTHCODE FIELD IS DROPPED)
        # 1 = MD2
        # 2 = MD5
        # 3 = reserved
        # 4 = straight password/key
        # 5 = Proprietary
                 
        # sequence number
        if seq == None:
            seq = 0

        # session ID
        if sid == None:
            sid = 0
                
        # Auth type
        if authtype == None:
            authtype = 0
            return struct.pack("<BLLB", authtype, seq, sid, leng)
        else:
            return struct.pack("<BLL16sB", authtype, seq, sid, mac, leng)

    # Construct IPMI source header.
    def build_ipmi_src_header(self, rqaddr, rqlun, rqseq, cmd):
        return struct.pack("<3b", rqaddr, ((rqseq << 2) + rqlun), cmd)

    # Construct IPMI destination header.
    def build_ipmi_dest_header(self, rsaddr, netfn, lun):
        # rsaddr appears to always be 0x20 for IPMIoLAN (BMC, lun 0).
        # The spec is unclear, and software IDs 0x40-0x46 (7-bit)
        # have been assigned to Remote Console 1-7. But there is a
        # Catch-22 .. how do you tell your ID when you need to put
        # your ID in your request?
        # solution ... use channel 0xe!
    
        # we have to combine these with their own checksum.
        nl = netfn << 2
        nl = nl + lun
        cksum = self.calc_ipmi_cksum("%c%c" % (rsaddr, nl))
        #print "header: rsaddr %d" % rsaddr
        #print "header: nl %d" % nl
        #print "header: cksum %d" % cksum
        return struct.pack("<3b", rsaddr, nl, cksum)

    # send_ipmi_command: Send IPMI command (including all headers and body)
    # to the configured peer. Slap in auth info if available.
    def send_ipmi_command(self, header, body):
        # calc cksum on body
        bck = self.calc_ipmi_cksum(body)

        # Composite the entire IPMI Message

        ipmimessage = header + body + struct.pack("<b", bck)

        # Calculate the AuthCode
        if self.authtype == IPMI_AUTHT_MD5:
            # for MD* auth, we composite
            # password, session ID, ipmimessage, seq#, password
            # password must be 0-padded to 16 bytes
            m = md5.new()
            m.update(struct.pack("<16s", self.password))
            m.update(struct.pack("<L", self.sessid))
            m.update(ipmimessage)
            m.update(struct.pack("<L", self.seq))
            m.update(struct.pack("<16s", self.password))
            self.authcode = m.digest()
        elif self.authtype == IPMI_AUTHT_MD2:
            raise ValueError, "MD2 Authentication is not supported"
        elif self.authtype == IPMI_AUTHT_PASSWD:
            # for password auth, just send the password (NO NO NO!!)
            self.authcode = self.password

        # Get the IPMI channel header, with message length.
        #print "Authtype = %s, seq = %x, sessid=%x" % (type(self.authtype), self.seq, self.sessid)
        ipmisess = self.build_ipmi_channel_header(self.authtype, self.seq, self.sessid, self.authcode, len(ipmimessage))

        # Glue in RMCP header.
        # XXX this should use hton* and be put in __init__
        # since it's static
        rmcphead = struct.pack('<4b', 0x06, 0x00, 0xff, 0x07)

        #print "SENDING: %s" % `(rmcphead + ipmisess + ipmimessage)`

        # send!
        self.sock.sendto(rmcphead + ipmisess + ipmimessage, (self.peer, 0x26f))

        # Increment seq#
        if self.seq != 0 and self.seq != None:
            self.seq = self.seq + 1

        # Wait around for response...
        resp = self.sock.recvfrom(100)
        return resp[0]

    # Extract the IPMI response information.
    def get_ipmi_body(self, resp):
        # first four bytes are the RMCP header, we can chuck 'em.
        # Next 1+8 are the session header
        #print "IPMI Response (%d): %s" % (len(resp), `resp`)
        ipmisess = resp[4:13]
        #print "IPMI session header: %s" % `ipmisess`
        # unpack!
        (authtype, seq, sid) = struct.unpack("<BLL", ipmisess)
        #print "IPMI Session Data: authtype=%#x, seq=%#x, sid=%#x" % (authtype, seq, sid)
        if authtype != 0x0 and authtype != 0x1:
            #print "** Authenticated packet **"
            # pull off the MAC (16 bytes)
            mac = resp[13:29]
            #print "** len(mac) = %d" % len(mac)
            # XXX must validate packet is authentic
            # next is length (what nut put this here? I smell overflow!)
            leng = ord(resp[29])
            #print "Read length %d" % leng
            # rest should be the message body, unless that stupid pad byte
            # is hiding in there.
            return resp[30:-1]
        else:
            leng = ord(resp[13])
            #print "NO AUTH: Length %d: Handing back %s" % (leng, `resp[14:]`)
            return resp[14:]

    # Extract the actual response data from the IPMI/IPMB packet.
    def get_response_data(self, body):
        (rqaddr, nl, cksum, rsaddr, rqsl) = struct.unpack("<5B", body[:5]);
        #print "IPMB HEADER: rqaddr %#x, netfn/rqlun %#x, cksum %#x,\nIPMB HEADER: rsaddr %#x, rqseq/rslun %#x" % (rqaddr, nl, cksum, rsaddr, rqsl)
        return body[6:]

    # big-ass macro for building & sending ipmi data
    def send(self, daddr, dlun, saddr, slun, sseq, netfn, cmd, data):
        header = self.build_ipmi_dest_header(daddr, netfn, dlun)
        body = self.build_ipmi_src_header(saddr, slun, sseq, cmd)
        body = body + data
        ret = self.send_ipmi_command(header, body)
        return self.get_response_data(self.get_ipmi_body(ret))

    #####################################################
    #  COMMAND IMPLEMENTATIONS
    #####################################################
        
    # Determine auth capabilities of the peer.
    def get_auth_cap(self):
        # request data is two bytes, first is the channel num (0xe == this),
        # second is the max priv level (0x4=admin)
       # Get Channel Authentication Capabilities is App/0x38
        header = self.build_ipmi_dest_header(0x20, 0x6, 0x0)
        # rqaddr, rqlun/seq, cmd, reqinfo
        # use our lun 0
        # body = struct.pack("<5b", 0x20, 0x01<<2, 0x38, 0x0e, 0x04)
        # Ask about channel #1 (serial)
        #                         daddr srclun   cmd   chan  max priv
        body = struct.pack("<5b", 0x20, 0x01<<2, 0x38, 0x0e, self.authlevel)
        resp = self.send_ipmi_command(header, body)
        #print "IPMI RESPONSE (%d): %s" % (len(resp), `resp`)
        ibody = self.get_ipmi_body(resp)
        rbody = self.get_response_data(ibody)

        # now we have our response, parse the puppy out.
        # This is in the Section 12.4 format, with the IPMB header
        # still stuffed on it.
        
        (ccode, chan, auths, aflags, reserved, oemid, oemdata) = struct.unpack("<5blb", rbody)
       ## print "Authentication Capabilities:"
##        print "Completion Code: 0x%x" % ccode
##        print "Channel Number : 0x%x" % chan
##        print "Auth Types     : 0x%x" % auths
        if auths & 0x1 != 0:
            #print "    None"
            self.authtypes.append('None')
        if auths & 0x2 != 0:
            #print "    MD2"
            self.authtypes.append('MD2')
        if auths & 0x4 != 0:
            #print "    MD5" 
            self.authtypes.append('MD5')
        # 0x8 is reserved
        if auths & 0x10 != 0:
            #print "    Password"
            self.authtypes.append('Password')
        if auths & 0x20 != 0:
            #print "    OEM Proprietary"
            self.authtypes.append('OEM')
##        print "Auth Flags     : 0x%x" % aflags
##        if aflags & 0x1 != 0:
##            print "    Anonymous Login enabled"
##        if aflags & 0x2 != 0:
##            print "    Null user name enabled"
##        if aflags & 0x4 != 0:
##            print "    Non-null usernames enabled"
##        if aflags & 0x8 != 0:
##            print "    User Level authentication disabled"
##        if aflags & 0x10 != 0:
##            print "    Per Message Authentication disabled"
##        print "IANA OEM ID    : 0x%x" % oemid
##        print "OEM Data       : 0x%x" % oemdata

    # Get initial session challenge with requested authtype
    def get_session_challenge(self, atype):
        # Get Session Challenge is App/0x39
    #                                        daddr netfn lun
        header = self.build_ipmi_dest_header(0x20, 0x6, 0x0)
        #print "self.authtype = %s, self.username = %s" % (atype, self.username)
        #                             daddr srclun   cmd     
        body = struct.pack("<4B16s", 0x20, 0x01<<2, 0x39,
                           # authtype       username
                           atype, self.username)
        resp = self.send_ipmi_command(header, body)
        #print "IPMI RESPONSE (%d): %s" % (len(resp), `resp`)
        ibody = self.get_ipmi_body(resp)
        rbody = self.get_response_data(ibody)
        # print "rbody (%d) = %s" % (len(rbody), `rbody`)
        (ccode, self.sessid, self.authcode) = struct.unpack("<BL16s", rbody[:21])
        # future packets must be marked with this
        self.authtype = atype
##        print "Session Challenge"
##        print "Completion Code: 0x%x" % ccode
##        print "Session ID     : %x" % self.sessid
##        print "Challenge      : %s" % `self.authcode`

    # Activate!
    def activate_session(self):
        # The fields for this get kinda tricky.  We start using the session
        # fields.  authtype is what we've been using; seq# is still 0;
        # sessid is the temporary one from GSC; we calcluate the authcode
        # for this message.
        # In the body, we re-pass the authtype, the max priv level,
        # the original challenge, and a remote ISN.

        # Activate Session is App/0x3a
        #                                   daddr netfn lun
        header = self.build_ipmi_dest_header(0x20, 0x6, 0x0)

        #                             daddr srclun   cmd

        # self.authcode is still the challenge at this point
        body = struct.pack("<5B16sL", 0x20, 0x01<<2, 0x3a,
                           # authtype, maxpriv, challenge, ISN
                           self.authtype, self.authlevel, self.authcode, 0x1)
        resp = self.send_ipmi_command(header, body)
        #print "IPMI RESPONSE (%d): %s" % (len(resp), `resp`)
        ibody = self.get_ipmi_body(resp)
        rbody = self.get_response_data(ibody)
        #print "rbody (%d) = %s" % (len(rbody), `rbody`)

        # well, if this worked, we should have something like this:
        try:
            (ccode, atype, self.sessid, self.seq, maxpriv) = struct.unpack("<2B2LB", rbody[:21])
        except:
            print "Error activating session (0x%x)" % ord(rbody[0])
            raise
##        print "Activate Session"
##        print "Completion Code: 0x%x" % ccode
##        print "Accepted AType : 0x%x" % atype
##        print "Session ID     : 0x%x" % self.sessid
##        print "Initial Seq #  : 0x%x" % self.seq
##        print "Max Priv Level : 0x%x" % maxpriv

    # Set authorization level. Since we start up in User, we need
    # to bump it up before doing anything else.
    def set_auth_level(self):
        ret = self.send(0x20, 0x0, 0x81, 0x0, 41, 0x6, 0x3b, struct.pack('<B', 4))
        if ord(ret[0]) != 0:
            raise ValueError, "Couldn't raise auth level"

    # Establish session with peer.
    def establish(self):
        # Bind our port.
        # Before we can bring the session up, we need to ask the host
        # what authentication is required/available.
        # get_auth_cap will load self.authtypes with the available
        # authentication methods.
        self.get_auth_cap()

        # Pick the best one.
        if 'MD5' in self.authtypes:
            atype = IPMI_AUTHT_MD5
        elif 'MD2' in self.authtypes:
            atype = IPMI_AUTHT_MD2
            print "*** MD2 Authentication is not supported! ***"
        elif 'Password' in self.authtypes:
            atype = IPMI_AUTHT_PASSWD
            print "*** DANGER: Using password auth ***"
            print "*** THE PASSWORD WILL BE SENT IN THE CLEAR! ***"
        else:
            raise exceptions.ValueError, "No supported auth types returned"

        # Get the initial session challenge.
        self.get_session_challenge(atype)

        # Activate session.
        self.activate_session()

        # Reset auth level.
        self.set_auth_level()

        # We are on-line!
        self.connected = 1
        
class IPMISystemInterface:
    def __init__(self, interface="SMIC"):
        # XXX
        import sys
        sys.path.append('ioport')
        import string
        import ioport

        self.interface = string.upper(interface)

        if self.interface not in ("KCS", "SMIC"):
            raise ValueError, "Not a valid interface"

        if self.interface == "KCS":
            import ipmikcs
            ioport.ioperm(ipmikcs.KCS_BASE_IOPORT, 2, 1)

        elif self.interface == "SMIC":
            import ipmismic
            ioport.ioperm(ipmismic.SMIC_BASE_PORT, 3, 1)

        else:
            print "huh?"


    # IPMISession-compatible send
    def send(self, daddr, dlun, saddr, slun, sseq, netfn, cmd, data):
        # in SI land, we only need the netfn, command, and data
        tx = ""
        tx = struct.pack('<BB', (netfn << 2 + slun), cmd)
        tx = tx + data

        if self.interface == "KCS":
            import ipmikcs
            ipmikcs.kcs_write(tx)
            return ipmikcs.kcs_read()[2:]

        if self.interface == "SMIC":
            import ipmismic
            ipmismic.smic_write(tx)
            return ipmismic.smic_read()[2:]

    def establish(self):
        pass
        
